@@ -30,18 +30,27 @@ pub const MAX_CACHE_BLOCK_SIZE: Sectors = Sectors(2 * IEC::Mi); // 1 GiB
 
 const CACHE_TARGET_NAME: &str = "cache";
 
+/// Struct representing params for a cache target
 #[derive(Clone, Debug, Eq, PartialEq)]
 pub struct CacheTargetParams {
+    /// Cache metadata device
     pub meta: Device,
+    /// Cache device
     pub cache: Device,
+    /// Origin device with data to be cached
     pub origin: Device,
+    /// Cache block size
     pub cache_block_size: Sectors,
+    /// Feature arguments
     pub feature_args: HashSet<String>,
+    /// IO policy
     pub policy: String,
+    /// IO policy arguments
     pub policy_args: HashMap<String, String>,
 }
 
 impl CacheTargetParams {
+    /// Create a new CacheTargetParams struct
     pub fn new(
         meta: Device,
         cache: Device,
@@ -154,7 +163,7 @@ impl TargetParams for CacheTargetParams {
                 self.policy_args.len(),
                 self.policy_args
                     .iter()
-                    .map(|(k, v)| format!("{} {}", k, v))
+                    .map(|(k, v)| format!("{k} {v}"))
                     .collect::<Vec<String>>()
                     .join(" ")
             )
@@ -177,12 +186,15 @@ impl TargetParams for CacheTargetParams {
     }
 }
 
+/// A target table for a cache device.
 #[derive(Clone, Debug, Eq, PartialEq)]
 pub struct CacheDevTargetTable {
+    /// The device's table
     pub table: TargetLine<CacheTargetParams>,
 }
 
 impl CacheDevTargetTable {
+    /// Make a new CacheDevTargetTable from the required input
     pub fn new(start: Sectors, length: Sectors, params: CacheTargetParams) -> CacheDevTargetTable {
         CacheDevTargetTable {
             table: TargetLine::new(start, length, params),
@@ -210,7 +222,7 @@ impl TargetTable for CacheDevTargetTable {
         Ok(CacheDevTargetTable::new(
             Sectors(line.0),
             Sectors(line.1),
-            format!("{} {}", line.2.to_string(), line.3).parse::<CacheTargetParams>()?,
+            format!("{} {}", line.2, line.3).parse::<CacheTargetParams>()?,
         ))
     }
 
@@ -556,12 +568,12 @@ impl CacheDev {
         cache_block_size: Sectors,
     ) -> DmResult<CacheDev> {
         if device_exists(dm, name)? {
-            let err_msg = format!("cachedev {} already exists", name);
+            let err_msg = format!("cachedev {name} already exists");
             return Err(DmError::Dm(ErrorEnum::Invalid, err_msg));
         }
 
         let table = CacheDev::gen_default_table(&meta, &cache, &origin, cache_block_size);
-        let dev_info = device_create(dm, name, uuid, &table, DmOptions::default())?;
+        let dev_info = device_create(dm, name, uuid, &table, DmOptions::private())?;
 
         Ok(CacheDev {
             dev_info: Box::new(dev_info),
@@ -595,7 +607,7 @@ impl CacheDev {
             device_match(dm, &dev, uuid)?;
             dev
         } else {
-            let dev_info = device_create(dm, name, uuid, &table, DmOptions::default())?;
+            let dev_info = device_create(dm, name, uuid, &table, DmOptions::private())?;
             CacheDev {
                 dev_info: Box::new(dev_info),
                 meta_dev: meta,
@@ -681,11 +693,15 @@ impl CacheDev {
 
     /// Generate a table to be passed to DM. The format of the table
     /// entries is:
+    /// ```plain
     /// <start sec (0)> <length> "cache" <cache-specific string>
+    /// ```
     /// where the cache-specific string has the format:
+    /// ```plain
     /// <meta maj:min> <cache maj:min> <origin maj:min> <block size>
     /// <#num feature args (1)> writethrough <replacement policy (default)>
     /// <#num policy args (0)>
+    /// ```
     /// There is exactly one entry in the table.
     /// Various defaults are hard coded in the method.
     fn gen_default_table(
@@ -713,7 +729,7 @@ impl CacheDev {
     fn parse_pairs(start_index: usize, vals: &[&str]) -> DmResult<(usize, Vec<(String, String)>)> {
         let num_pairs: usize = parse_value(vals[start_index], "number of pairs")?;
         if num_pairs % 2 != 0 {
-            let err_msg = format!("Number of args \"{}\" is not even", num_pairs);
+            let err_msg = format!("Number of args \"{num_pairs}\" is not even");
             return Err(DmError::Dm(ErrorEnum::Invalid, err_msg));
         }
         let next_start_index = start_index + num_pairs + 1;
@@ -829,18 +845,12 @@ mod tests {
                     cache.table.table.params.cache_block_size
                 );
 
-                // No data means no cache blocks used
-                assert_eq!(usage.used_cache, DataBlocks(0));
-
                 let performance = &status.performance;
 
-                // No activity should mean all performance data is 0
-                assert_eq!(performance.read_hits, 0);
-                assert_eq!(performance.read_misses, 0);
+                // No write activity should mean all write performance data is 0
                 assert_eq!(performance.write_hits, 0);
                 assert_eq!(performance.write_misses, 0);
                 assert_eq!(performance.demotions, 0);
-                assert_eq!(performance.promotions, 0);
                 assert_eq!(performance.dirty, 0);
 
                 // The current defaults for configuration values
@@ -857,7 +867,7 @@ mod tests {
 
                 assert!(!status.needs_check);
             }
-            status => panic!("unexpected thinpool status: {:?}", status),
+            status => panic!("unexpected thinpool status: {status:?}"),
         }
 
         let table = CacheDev::read_kernel_table(&dm, &DevId::Name(cache.name()))
