@@ -2,13 +2,14 @@
 // License, v. 2.0. If a copy of the MPL was not distributed with this
 // file, You can obtain one at http://mozilla.org/MPL/2.0/.
 
-use std::{fmt, io, path::Path, str::FromStr};
+use std::{fmt, path::Path, str::FromStr};
 
+use cfg_if::cfg_if;
 use nix::libc::{dev_t, major, makedev, minor};
 use nix::sys::stat::{self, SFlag};
 
 use crate::{
-    core::errors::ErrorKind,
+    core::errors,
     result::{DmError, DmResult},
 };
 
@@ -23,7 +24,7 @@ pub struct Device {
     pub minor: u32,
 }
 
-/// Display format is the device number in "<major>:<minor>" format
+/// Display format is the device number in `<major>:<minor>` format
 impl fmt::Display for Device {
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         write!(f, "{}:{}", self.major, self.minor)
@@ -36,26 +37,20 @@ impl FromStr for Device {
     fn from_str(s: &str) -> Result<Device, DmError> {
         let vals = s.split(':').collect::<Vec<_>>();
         if vals.len() != 2 {
-            let err_msg = format!("value \"{}\" split into wrong number of fields", s);
-            return Err(DmError::Core(ErrorKind::InvalidArgument(err_msg).into()));
+            let err_msg = format!("value \"{s}\" split into wrong number of fields");
+            return Err(DmError::Core(errors::Error::InvalidArgument(err_msg)));
         }
         let major = vals[0].parse::<u32>().map_err(|_| {
-            DmError::Core(
-                ErrorKind::InvalidArgument(format!(
-                    "could not parse \"{}\" to obtain major number",
-                    vals[0]
-                ))
-                .into(),
-            )
+            DmError::Core(errors::Error::InvalidArgument(format!(
+                "could not parse \"{}\" to obtain major number",
+                vals[0]
+            )))
         })?;
         let minor = vals[1].parse::<u32>().map_err(|_| {
-            DmError::Core(
-                ErrorKind::InvalidArgument(format!(
-                    "could not parse \"{}\" to obtain minor number",
-                    vals[1]
-                ))
-                .into(),
-            )
+            DmError::Core(errors::Error::InvalidArgument(format!(
+                "could not parse \"{}\" to obtain minor number",
+                vals[1]
+            )))
         })?;
         Ok(Device { major, minor })
     }
@@ -63,10 +58,12 @@ impl FromStr for Device {
 
 impl From<dev_t> for Device {
     fn from(val: dev_t) -> Device {
+        #[allow(unused_unsafe)]
         let major = unsafe { major(val) };
         #[cfg(target_os = "android")]
         let major = major as u32;
 
+        #[allow(unused_unsafe)]
         let minor = unsafe { minor(val) };
         #[cfg(target_os = "android")]
         let minor = minor as u32;
@@ -77,13 +74,21 @@ impl From<dev_t> for Device {
 
 impl From<Device> for dev_t {
     fn from(dev: Device) -> dev_t {
-        #[cfg(target_os = "android")]
-        unsafe {
-            makedev(dev.major as i32, dev.minor as i32)
-        }
-        #[cfg(not(target_os = "android"))]
-        unsafe {
-            makedev(dev.major, dev.minor)
+        cfg_if! {
+            if #[cfg(target_os = "android")] {
+                // dev.major, dev.minor u32 in libc 0.2.133
+                #[allow(clippy::useless_conversion)]
+                makedev(
+                    dev.major
+                        .try_into()
+                        .expect("value is smaller than max positive i32"),
+                    dev.minor
+                        .try_into()
+                        .expect("value is smaller than max positive i32"),
+                )
+            } else {
+                makedev(dev.major, dev.minor)
+            }
         }
     }
 }
@@ -122,11 +127,11 @@ pub fn devnode_to_devno(path: &Path) -> DmResult<Option<u64>> {
                 None
             },
         ),
-        Err(err) if err == nix::Error::ENOENT => Ok(None),
-        Err(err) => Err(DmError::Core(
-            ErrorKind::MetadataIoError(path.to_owned(), io::Error::new(io::ErrorKind::Other, err))
-                .into(),
-        )),
+        Err(nix::Error::ENOENT) => Ok(None),
+        Err(err) => Err(DmError::Core(errors::Error::MetadataIo(
+            path.to_owned(),
+            err.to_string(),
+        ))),
     }
 }
 
