@@ -4,7 +4,6 @@
 
 use std::{
     cmp,
-    convert::TryInto,
     fs::File,
     io::{Cursor, Read, Write},
     mem::size_of,
@@ -12,7 +11,9 @@ use std::{
     slice, str,
 };
 
-use nix::libc::ioctl as nix_ioctl;
+use nix::{errno, libc::ioctl as nix_ioctl};
+use retry::{delay::Fixed, retry_with_index, Error as RetryError, OperationResult};
+use semver::Version;
 
 use crate::{
     core::{
@@ -21,6 +22,7 @@ use crate::{
         dm_flags::DmFlags,
         dm_ioctl as dmi,
         dm_options::DmOptions,
+        dm_udev_sync::{UdevSync, UdevSyncAction},
         errors,
         types::{DevId, DmName, DmNameBuf, DmUuid},
         util::{
@@ -41,6 +43,12 @@ const DM_CTL_PATH: &str = "/dev/device-mapper";
 /// Start with a large buffer to make BUFFER_FULL rare. Libdm does this too.
 const MIN_BUF_SIZE: usize = 16 * 1024;
 
+/// Number of device remove retry attempts
+const DM_REMOVE_RETRIES: usize = 5;
+
+/// Delay between remove attempts
+const DM_REMOVE_MSLEEP_DELAY: u64 = 200;
+
 /// Context needed for communicating with devicemapper.
 pub struct DM {
     file: File,
@@ -54,17 +62,13 @@ impl DmOptions {
         allowable_flags: DmFlags,
     ) -> DmResult<dmi::Struct_dm_ioctl> {
         let clean_flags = allowable_flags & self.flags();
-        let event_nr = u32::from(self.cookie().bits()) << 16;
-        let mut hdr: dmi::Struct_dm_ioctl = Default::default();
-
-        hdr.version[0] = dmi::DM_VERSION_MAJOR;
-        hdr.version[1] = dmi::DM_VERSION_MINOR;
-        hdr.version[2] = dmi::DM_VERSION_PATCHLEVEL;
-
-        hdr.flags = clean_flags.bits();
-        hdr.event_nr = event_nr;
-
-        hdr.data_start = size_of::<dmi::Struct_dm_ioctl>() as u32;
+        let event_nr = self.udev_flags().bits() << dmi::DM_UDEV_FLAGS_SHIFT;
+        let mut hdr: dmi::Struct_dm_ioctl = devicemapper_sys::dm_ioctl {
+            flags: clean_flags.bits(),
+            event_nr,
+            data_start: size_of::<dmi::Struct_dm_ioctl>() as u32,
+            ..Default::default()
+        };
 
         if let Some(id) = id {
             match id {
@@ -107,109 +111,123 @@ impl DM {
         &self.file
     }
 
-    // Give this a filled-in header and optionally add'l stuff.
-    // Does the ioctl and maybe returns stuff. Handles BUFFER_FULL flag.
-    //
+    // Make the ioctl call specified by the given ioctl number.
+    // Set the required DM version to the lowest that supports the given ioctl.
     fn do_ioctl(
         &self,
         ioctl: u8,
         hdr: &mut dmi::Struct_dm_ioctl,
         in_data: Option<&[u8]>,
-    ) -> DmResult<Vec<u8>> {
-        // Create in-buf by copying hdr and any in-data into a linear
-        // Vec v.  'hdr_slc' also aliases hdr as a &[u8], used first
-        // to copy the hdr into v, and later to update the
-        // possibly-modified hdr.
+    ) -> DmResult<(DeviceInfo, Vec<u8>)> {
+        let op = request_code_readwrite!(dmi::DM_IOCTL, ioctl, size_of::<dmi::Struct_dm_ioctl>());
+        #[cfg(target_os = "android")]
+        let op = op as i32;
 
-        // Start with a large buffer to make BUFFER_FULL rare. Libdm
-        // does this too.
-        hdr.data_size = cmp::max(
+        let ioctl_version = dmi::ioctl_to_version(ioctl);
+        hdr.version[0] = ioctl_version.0;
+        hdr.version[1] = ioctl_version.1;
+        hdr.version[2] = ioctl_version.2;
+
+        // Begin udev sync transaction and set DM_UDEV_PRIMARY_SOURCE_FLAG
+        // if ioctl command generates uevents.
+        let sync = UdevSync::begin(hdr, ioctl)?;
+
+        let data_size = cmp::max(
             MIN_BUF_SIZE,
             size_of::<dmi::Struct_dm_ioctl>() + in_data.map_or(0, |x| x.len()),
-        ) as u32;
-        let mut v: Vec<u8> = Vec::with_capacity(hdr.data_size as usize);
+        );
 
-        let hdr_slc = unsafe {
-            let len = hdr.data_start as usize;
-            let ptr = hdr as *mut dmi::Struct_dm_ioctl as *mut u8;
-            slice::from_raw_parts_mut(ptr, len)
-        };
-
-        v.extend_from_slice(hdr_slc);
-        if let Some(in_data) = in_data {
-            v.extend(in_data.iter().cloned());
-        }
-
-        // zero out the rest
-        let cap = v.capacity();
-        v.resize(cap, 0);
-        let mut hdr = unsafe {
-            (v.as_mut_ptr() as *mut dmi::Struct_dm_ioctl)
-                .as_mut()
-                .expect("pointer to own structure v can not be NULL")
-        };
-        let op = request_code_readwrite!(dmi::DM_IOCTL, ioctl, size_of::<dmi::Struct_dm_ioctl>());
+        let mut buffer: Vec<u8> = Vec::with_capacity(data_size);
+        let mut buffer_hdr;
         loop {
-            #[cfg(target_os = "android")]
-            let op = op as i32;
-            if let Err(err) =
-                unsafe { convert_ioctl_res!(nix_ioctl(self.file.as_raw_fd(), op, v.as_mut_ptr())) }
-            {
+            hdr.data_size = buffer.capacity() as u32;
+
+            let hdr_slc = unsafe {
+                let len = hdr.data_start as usize;
+                let ptr = hdr as *mut dmi::Struct_dm_ioctl as *mut u8;
+                slice::from_raw_parts_mut(ptr, len)
+            };
+
+            buffer.clear();
+            buffer.extend_from_slice(hdr_slc);
+            if let Some(in_data) = in_data {
+                buffer.extend(in_data.iter().cloned());
+            }
+            buffer.resize(buffer.capacity(), 0);
+
+            buffer_hdr = unsafe { &mut *(buffer.as_mut_ptr() as *mut dmi::Struct_dm_ioctl) };
+
+            if let Err(err) = unsafe {
+                convert_ioctl_res!(nix_ioctl(self.file.as_raw_fd(), op, buffer.as_mut_ptr()))
+            } {
+                // Cancel udev sync and clean up semaphore
+                sync.cancel();
                 return Err(DmError::Core(errors::Error::Ioctl(
+                    op as u8,
                     DeviceInfo::new(*hdr).ok().map(Box::new),
+                    DeviceInfo::new(*buffer_hdr).ok().map(Box::new),
                     Box::new(err),
                 )));
             }
-            // If DM was able to write the requested data into the provided buffer, break the loop
-            if (hdr.flags & DmFlags::DM_BUFFER_FULL.bits()) == 0 {
+
+            if (buffer_hdr.flags & DmFlags::DM_BUFFER_FULL.bits()) == 0 {
                 break;
             }
 
             // If DM_BUFFER_FULL is set, DM requires more space for the
-            // response.  Double the size of the buffer and re-try the ioctl.
-            // If the size of the buffer is already as large as can be possibly
-            // expressed in hdr.data_size field, return an error. Never allow
-            // the size to exceed u32::MAX.
-            let len = v.len();
+            // response.  Double the capacity of the buffer and re-try the
+            // ioctl. If the size of the buffer is already as large as can be
+            // possibly expressed in data_size field, return an error.
+            // Never allow the size to exceed u32::MAX.
+            let len = buffer.capacity();
             if len == u32::MAX as usize {
                 return Err(DmError::Core(errors::Error::IoctlResultTooLarge));
             }
-            v.resize((len as u32).saturating_mul(2) as usize, 0);
-
-            // v.resize() may move the buffer if the requested increase doesn't fit in continuous
-            // memory.  Update hdr to the possibly new address.
-            hdr = unsafe {
-                (v.as_mut_ptr() as *mut dmi::Struct_dm_ioctl)
-                    .as_mut()
-                    .expect("pointer to own structure v can not be NULL")
-            };
-            hdr.data_size = v.len() as u32;
+            buffer.resize((len as u32).saturating_mul(2) as usize, 0);
         }
 
-        // hdr possibly modified so copy back
-        hdr_slc.clone_from_slice(&v[..hdr.data_start as usize]);
+        let data_end = cmp::max(buffer_hdr.data_size, buffer_hdr.data_start);
 
-        // Return header data section.
-        let new_data_off = cmp::max(hdr.data_start, hdr.data_size);
-        Ok(v[hdr.data_start as usize..new_data_off as usize].to_vec())
+        // Synchronize with udev event processing
+        sync.end(buffer_hdr.flags)?;
+        Ok((
+            DeviceInfo::try_from(*buffer_hdr)?,
+            buffer[buffer_hdr.data_start as usize..data_end as usize].to_vec(),
+        ))
     }
 
     /// Devicemapper version information: Major, Minor, and patchlevel versions.
     pub fn version(&self) -> DmResult<(u32, u32, u32)> {
         let mut hdr = DmOptions::default().to_ioctl_hdr(None, DmFlags::empty())?;
 
-        self.do_ioctl(dmi::DM_VERSION_CMD as u8, &mut hdr, None)?;
+        let (hdr_out, _) = self.do_ioctl(dmi::DM_VERSION_CMD as u8, &mut hdr, None)?;
 
-        Ok((hdr.version[0], hdr.version[1], hdr.version[2]))
+        Ok((
+            hdr_out
+                .version()
+                .major
+                .try_into()
+                .expect("dm_ioctl struct field is u32"),
+            hdr_out
+                .version()
+                .minor
+                .try_into()
+                .expect("dm_ioctl struct field is u32"),
+            hdr_out
+                .version()
+                .patch
+                .try_into()
+                .expect("dm_ioctl struct field is u32"),
+        ))
     }
 
     /// Remove all DM devices and tables. Use discouraged other than
     /// for debugging.
     ///
-    /// If DM_DEFERRED_REMOVE is set, the request will succeed for
+    /// If `DM_DEFERRED_REMOVE` is set, the request will succeed for
     /// in-use devices, and they will be removed when released.
     ///
-    /// Valid flags: DM_DEFERRED_REMOVE
+    /// Valid flags: `DM_DEFERRED_REMOVE`
     pub fn remove_all(&self, options: DmOptions) -> DmResult<()> {
         let mut hdr = options.to_ioctl_hdr(None, DmFlags::DM_DEFERRED_REMOVE)?;
 
@@ -223,7 +241,9 @@ impl DM {
     /// support it, each device's last event_nr.
     pub fn list_devices(&self) -> DmResult<Vec<(DmNameBuf, Device, Option<u32>)>> {
         let mut hdr = DmOptions::default().to_ioctl_hdr(None, DmFlags::empty())?;
-        let data_out = self.do_ioctl(dmi::DM_LIST_DEVICES_CMD as u8, &mut hdr, None)?;
+        let (hdr_out, data_out) = self.do_ioctl(dmi::DM_LIST_DEVICES_CMD as u8, &mut hdr, None)?;
+
+        let event_nr_set = hdr_out.version() >= &Version::new(4, 37, 0);
 
         let mut devs = Vec::new();
         if !data_out.is_empty() {
@@ -241,7 +261,7 @@ impl DM {
                     (device.name.as_ptr() as *const u8).offset_from(device as *const _ as *const u8)
                 } as usize;
 
-                let dm_name = str_from_byte_slice(&result[name_offset as usize..])
+                let dm_name = str_from_byte_slice(&result[name_offset..])
                     .map(|s| s.to_owned())
                     .ok_or_else(|| {
                         DmError::Dm(
@@ -254,27 +274,23 @@ impl DM {
                 // DM version supports it.
                 // Should match offset calc in kernel's
                 // drivers/md/dm-ioctl.c:list_devices
-                let event_nr = {
-                    match hdr.version[1] {
-                        0..=36 => None,
-                        _ => {
-                            // offsetof "name" in Struct_dm_name_list.
-                            let offset =
-                                align_to(name_offset + dm_name.len() + 1, size_of::<u64>());
-                            let nr = u32::from_ne_bytes(
-                                result[offset..offset + size_of::<u32>()]
-                                    .try_into()
-                                    .map_err(|_| {
-                                        DmError::Dm(
-                                            ErrorEnum::Invalid,
-                                            "Incorrectly sized slice for u32".to_string(),
-                                        )
-                                    })?,
-                            );
+                let event_nr = if event_nr_set {
+                    // offsetof "name" in Struct_dm_name_list.
+                    let offset = align_to(name_offset + dm_name.len() + 1, size_of::<u64>());
+                    let nr = u32::from_ne_bytes(
+                        result[offset..offset + size_of::<u32>()]
+                            .try_into()
+                            .map_err(|_| {
+                                DmError::Dm(
+                                    ErrorEnum::Invalid,
+                                    "Incorrectly sized slice for u32".to_string(),
+                                )
+                            })?,
+                    );
 
-                            Some(nr)
-                        }
-                    }
+                    Some(nr)
+                } else {
+                    None
                 };
 
                 devs.push((DmNameBuf::new(dm_name)?, device.dev.into(), event_nr));
@@ -292,7 +308,7 @@ impl DM {
 
     /// Create a DM device. It starts out in a "suspended" state.
     ///
-    /// Valid flags: DM_READONLY, DM_PERSISTENT_DEV
+    /// Valid flags: `DM_READONLY`, `DM_PERSISTENT_DEV`
     ///
     /// # Example
     ///
@@ -319,32 +335,70 @@ impl DM {
             Self::hdr_set_uuid(&mut hdr, uuid)?;
         }
 
-        self.do_ioctl(dmi::DM_DEV_CREATE_CMD as u8, &mut hdr, None)?;
+        debug!("Creating device {name} (uuid={uuid:?})");
+        self.do_ioctl(dmi::DM_DEV_CREATE_CMD as u8, &mut hdr, None)
+            .map(|(hdr, _)| hdr)
+    }
 
-        DeviceInfo::new(hdr)
+    fn try_device_remove(
+        &self,
+        id: &DevId<'_>,
+        options: DmOptions,
+    ) -> OperationResult<DeviceInfo, DmError> {
+        let mut hdr = match options.to_ioctl_hdr(Some(id), DmFlags::DM_DEFERRED_REMOVE) {
+            Ok(hdr) => hdr,
+            Err(err) => {
+                return OperationResult::Err(err);
+            }
+        };
+
+        match self.do_ioctl(dmi::DM_DEV_REMOVE_CMD as u8, &mut hdr, None) {
+            Err(err) => {
+                if let DmError::Core(errors::Error::Ioctl(op, hdr_in, hdr_out, errno)) = err {
+                    if *errno == errno::Errno::EBUSY {
+                        OperationResult::Retry(DmError::Core(errors::Error::Ioctl(
+                            op, hdr_in, hdr_out, errno,
+                        )))
+                    } else {
+                        OperationResult::Err(DmError::Core(errors::Error::Ioctl(
+                            op, hdr_in, hdr_out, errno,
+                        )))
+                    }
+                } else {
+                    OperationResult::Err(err)
+                }
+            }
+            Ok((deviceinfo, _)) => OperationResult::Ok(deviceinfo),
+        }
     }
 
     /// Remove a DM device and its mapping tables.
     ///
-    /// If DM_DEFERRED_REMOVE is set, the request for an in-use
+    /// If `DM_DEFERRED_REMOVE` is set, the request for an in-use
     /// devices will succeed, and it will be removed when no longer
     /// used.
     ///
-    /// Valid flags: DM_DEFERRED_REMOVE
+    /// Valid flags: `DM_DEFERRED_REMOVE`
     pub fn device_remove(&self, id: &DevId<'_>, options: DmOptions) -> DmResult<DeviceInfo> {
-        let mut hdr = options.to_ioctl_hdr(Some(id), DmFlags::DM_DEFERRED_REMOVE)?;
-
-        self.do_ioctl(dmi::DM_DEV_REMOVE_CMD as u8, &mut hdr, None)?;
-
-        DeviceInfo::new(hdr)
+        debug!("Removing device {id}");
+        match retry_with_index(
+            Fixed::from_millis(DM_REMOVE_MSLEEP_DELAY).take(DM_REMOVE_RETRIES - 1),
+            |i| {
+                trace!("Device remove attempt {i} of {DM_REMOVE_RETRIES}");
+                self.try_device_remove(id, options)
+            },
+        ) {
+            Ok(deviceinfo) => Ok(deviceinfo),
+            Err(RetryError { error, .. }) => Err(error),
+        }
     }
 
     /// Change a DM device's name OR set the device's uuid for the first time.
     ///
-    /// Prerequisite: if new == DevId::Name(new_name), old_name != new_name
-    /// Prerequisite: if new == DevId::Uuid(uuid), device's current uuid
-    /// must be "".
-    /// Note: Possibly surprisingly, returned DeviceInfo's uuid or name field
+    /// Prerequisite: if `new == DevId::Name(new_name)`, `old_name != new_name`
+    /// Prerequisite: if `new == DevId::Uuid(uuid)`, device's current uuid
+    /// must be `""`.
+    /// Note: Possibly surprisingly, returned `DeviceInfo`'s uuid or name field
     /// contains the previous value, not the newly set value.
     pub fn device_rename(&self, old_name: &DmName, new: &DevId<'_>) -> DmResult<DeviceInfo> {
         let (options, id_in) = match *new {
@@ -355,28 +409,28 @@ impl DM {
             ),
         };
 
-        let data_in = [id_in, &[b'\0']].concat();
+        let data_in = [id_in, b"\0"].concat();
 
         let mut hdr = options.to_ioctl_hdr(None, DmFlags::DM_UUID)?;
         Self::hdr_set_name(&mut hdr, old_name)?;
 
-        self.do_ioctl(dmi::DM_DEV_RENAME_CMD as u8, &mut hdr, Some(&data_in))?;
-
-        DeviceInfo::new(hdr)
+        debug!("Renaming device {old_name} to {new}");
+        self.do_ioctl(dmi::DM_DEV_RENAME_CMD as u8, &mut hdr, Some(&data_in))
+            .map(|(hdr, _)| hdr)
     }
 
-    /// Suspend or resume a DM device, depending on if DM_SUSPEND flag
+    /// Suspend or resume a DM device, depending on if `DM_SUSPEND` flag
     /// is set or not.
     ///
     /// Resuming a DM device moves a table loaded into the "inactive"
-    /// slot by `table_load()` into the "active" slot.
+    /// slot by [`Self::table_load`] into the "active" slot.
     ///
     /// Will block until pending I/O is completed unless DM_NOFLUSH
     /// flag is given. Will freeze filesystem unless DM_SKIP_LOCKFS
     /// flags is given. Additional I/O to a suspended device will be
     /// held until it is resumed.
     ///
-    /// Valid flags: DM_SUSPEND, DM_NOFLUSH, DM_SKIP_LOCKFS
+    /// Valid flags: `DM_SUSPEND`, `DM_NOFLUSH`, `DM_SKIP_LOCKFS`
     ///
     /// # Example
     ///
@@ -394,9 +448,14 @@ impl DM {
             DmFlags::DM_SUSPEND | DmFlags::DM_NOFLUSH | DmFlags::DM_SKIP_LOCKFS,
         )?;
 
-        self.do_ioctl(dmi::DM_DEV_SUSPEND_CMD as u8, &mut hdr, None)?;
-
-        DeviceInfo::new(hdr)
+        let action = if options.flags().contains(DmFlags::DM_SUSPEND) {
+            "Suspending"
+        } else {
+            "Resuming"
+        };
+        debug!("{action} device {id}");
+        self.do_ioctl(dmi::DM_DEV_SUSPEND_CMD as u8, &mut hdr, None)
+            .map(|(hdr, _)| hdr)
     }
 
     /// Get DeviceInfo for a device. This is also returned by other
@@ -405,15 +464,15 @@ impl DM {
     pub fn device_info(&self, id: &DevId<'_>) -> DmResult<DeviceInfo> {
         let mut hdr = DmOptions::default().to_ioctl_hdr(Some(id), DmFlags::empty())?;
 
-        self.do_ioctl(dmi::DM_DEV_STATUS_CMD as u8, &mut hdr, None)?;
-
-        DeviceInfo::new(hdr)
+        trace!("Retrieving info for {id}");
+        self.do_ioctl(dmi::DM_DEV_STATUS_CMD as u8, &mut hdr, None)
+            .map(|(hdr, _)| hdr)
     }
 
     /// Wait for a device to report an event.
     ///
     /// Once an event occurs, this function behaves just like
-    /// `table_status`, see that function for more details.
+    /// [`Self::table_status`], see that function for more details.
     ///
     /// This interface is not very friendly to monitoring multiple devices.
     /// Events are also exported via uevents, that method may be preferable.
@@ -425,23 +484,19 @@ impl DM {
     ) -> DmResult<(DeviceInfo, Vec<(u64, u64, String, String)>)> {
         let mut hdr = options.to_ioctl_hdr(Some(id), DmFlags::DM_QUERY_INACTIVE_TABLE)?;
 
-        let data_out = self.do_ioctl(dmi::DM_DEV_WAIT_CMD as u8, &mut hdr, None)?;
+        trace!("Waiting on event for {id}");
+        let (hdr_out, data_out) = self.do_ioctl(dmi::DM_DEV_WAIT_CMD as u8, &mut hdr, None)?;
 
         let status = DM::parse_table_status(hdr.target_count, &data_out)?;
 
-        DeviceInfo::new(hdr).map(|info| (info, status))
+        Ok((hdr_out, status))
     }
 
     /// Load targets for a device into its inactive table slot.
     ///
-    /// `targets` is an array of (sector_start, sector_length, type, params).
+    /// `targets` is an array of `(sector_start, sector_length, type, params)`.
     ///
-    /// `params` are target-specific, please see [Linux kernel documentation]
-    /// https://git.kernel.org/cgit/linux/kernel/git/torvalds/linux.git/tree/ ->
-    /// Documentation/device-mapper
-    /// for more.
-    ///
-    /// `options` Valid flags: DM_READ_ONLY, DM_SECURE_DATA
+    /// `options` Valid flags: `DM_READ_ONLY`, `DM_SECURE_DATA`
     ///
     /// # Example
     ///
@@ -517,18 +572,18 @@ impl DM {
         // Flatten targets into a buf
         let data_in = cursor.into_inner();
 
-        self.do_ioctl(dmi::DM_TABLE_LOAD_CMD as u8, &mut hdr, Some(&data_in))?;
-
-        DeviceInfo::new(hdr)
+        trace!("Loading table \"{targets:?}\" for {id}");
+        self.do_ioctl(dmi::DM_TABLE_LOAD_CMD as u8, &mut hdr, Some(&data_in))
+            .map(|(hdr, _)| hdr)
     }
 
     /// Clear the "inactive" table for a device.
     pub fn table_clear(&self, id: &DevId<'_>) -> DmResult<DeviceInfo> {
         let mut hdr = DmOptions::default().to_ioctl_hdr(Some(id), DmFlags::empty())?;
 
-        self.do_ioctl(dmi::DM_TABLE_CLEAR_CMD as u8, &mut hdr, None)?;
-
-        DeviceInfo::new(hdr)
+        trace!("Clearing inactive table for {id}");
+        self.do_ioctl(dmi::DM_TABLE_CLEAR_CMD as u8, &mut hdr, None)
+            .map(|(hdr, _)| hdr)
     }
 
     /// Query DM for which devices are referenced by the "active"
@@ -541,17 +596,14 @@ impl DM {
     pub fn table_deps(&self, id: &DevId<'_>, options: DmOptions) -> DmResult<Vec<Device>> {
         let mut hdr = options.to_ioctl_hdr(Some(id), DmFlags::DM_QUERY_INACTIVE_TABLE)?;
 
-        let data_out = self.do_ioctl(dmi::DM_TABLE_DEPS_CMD as u8, &mut hdr, None)?;
+        trace!("Querying dependencies for {id}");
+        let (_, data_out) = self.do_ioctl(dmi::DM_TABLE_DEPS_CMD as u8, &mut hdr, None)?;
 
         if data_out.is_empty() {
             Ok(vec![])
         } else {
             let result = &data_out[..];
-            let target_deps = unsafe {
-                (result.as_ptr() as *const dmi::Struct_dm_target_deps)
-                    .as_ref()
-                    .expect("pointer to own structure result can not be NULL")
-            };
+            let target_deps = unsafe { &*(result.as_ptr() as *const dmi::Struct_dm_target_deps) };
 
             let dev_slc = unsafe {
                 slice::from_raw_parts(
@@ -571,17 +623,10 @@ impl DM {
 
     /// Parse a device's table. The table value is in buf, count indicates the
     /// expected number of lines.
-    /// Panics if there is an error parsing the table.
     /// Trims trailing white space off final entry on each line. This
     /// canonicalization makes checking identity of tables easier.
     /// Postcondition: The length of the next to last entry in any tuple is
     /// no more than 16 characters.
-    // Justification: If the ioctl succeeded, the data is correct and
-    // complete. An error in parsing can only result from a change in the
-    // kernel. We rely on DM's interface versioning system. Kernel changes
-    // will either be backwards-compatible, or will increment
-    // DM_VERSION_MAJOR.  Since calls made with a non-matching major version
-    // will fail, this protects against callers parsing unknown formats.
     fn parse_table_status(count: u32, buf: &[u8]) -> DmResult<Vec<(u64, u64, String, String)>> {
         let mut targets = Vec::new();
         if !buf.is_empty() {
@@ -589,11 +634,7 @@ impl DM {
 
             for _ in 0..count {
                 let result = &buf[next_off..];
-                let targ = unsafe {
-                    (result.as_ptr() as *const dmi::Struct_dm_target_spec)
-                        .as_ref()
-                        .expect("assume all parsing succeeds")
-                };
+                let targ = unsafe { &*(result.as_ptr() as *const dmi::Struct_dm_target_spec) };
 
                 let target_type = str_from_c_str(&targ.target_type)
                     .ok_or_else(|| {
@@ -661,30 +702,29 @@ impl DM {
             DmFlags::DM_NOFLUSH | DmFlags::DM_STATUS_TABLE | DmFlags::DM_QUERY_INACTIVE_TABLE,
         )?;
 
-        let data_out = self.do_ioctl(dmi::DM_TABLE_STATUS_CMD as u8, &mut hdr, None)?;
+        trace!("Retrieving table status for {id}");
+        let (hdr_out, data_out) = self.do_ioctl(dmi::DM_TABLE_STATUS_CMD as u8, &mut hdr, None)?;
 
-        let status = DM::parse_table_status(hdr.target_count, &data_out)?;
+        let status = DM::parse_table_status(hdr_out.target_count, &data_out)?;
 
-        DeviceInfo::new(hdr).map(|info| (info, status))
+        Ok((hdr_out, status))
     }
 
     /// Returns a list of each loaded target type with its name, and
     /// version broken into major, minor, and patchlevel.
+    #[cfg(devicemapper41supported)]
     pub fn list_versions(&self) -> DmResult<Vec<(String, u32, u32, u32)>> {
         let mut hdr = DmOptions::default().to_ioctl_hdr(None, DmFlags::empty())?;
 
-        let data_out = self.do_ioctl(dmi::DM_LIST_VERSIONS_CMD as u8, &mut hdr, None)?;
+        trace!("Listing loaded target versions");
+        let (_, data_out) = self.do_ioctl(dmi::DM_LIST_VERSIONS_CMD as u8, &mut hdr, None)?;
 
         let mut targets = Vec::new();
         if !data_out.is_empty() {
             let mut result = &data_out[..];
 
             loop {
-                let tver = unsafe {
-                    (result.as_ptr() as *const dmi::Struct_dm_target_versions)
-                        .as_ref()
-                        .expect("pointer to own structure result can not be NULL")
-                };
+                let tver = unsafe { &*(result.as_ptr() as *const dmi::Struct_dm_target_versions) };
 
                 let name =
                     str_from_byte_slice(&result[size_of::<dmi::Struct_dm_target_versions>()..])
@@ -711,6 +751,7 @@ impl DM {
     /// Send a message to the device specified by id and the sector
     /// specified by sector. If sending to the whole device, set sector to
     /// None.
+    #[cfg(devicemapper42supported)]
     pub fn target_msg(
         &self,
         id: &DevId<'_>,
@@ -731,9 +772,11 @@ impl DM {
         data_in.extend(msg.as_bytes());
         data_in.push(b'\0');
 
-        let data_out = self.do_ioctl(dmi::DM_TARGET_MSG_CMD as u8, &mut hdr, Some(&data_in))?;
+        debug!("Sending target message \"{msg}\" to {id}");
+        let (hdr_out, data_out) =
+            self.do_ioctl(dmi::DM_TARGET_MSG_CMD as u8, &mut hdr, Some(&data_in))?;
 
-        let output = if (hdr.flags & DmFlags::DM_DATA_OUT.bits()) > 0 {
+        let output = if (hdr_out.flags().bits() & DmFlags::DM_DATA_OUT.bits()) > 0 {
             Some(
                 str::from_utf8(&data_out[..data_out.len() - 1])
                     .map(|res| res.to_string())
@@ -747,7 +790,7 @@ impl DM {
         } else {
             None
         };
-        DeviceInfo::new(hdr).map(|info| (info, output))
+        Ok((hdr_out, output))
     }
 
     /// If DM is being used to poll for events, once it indicates readiness it
@@ -757,9 +800,9 @@ impl DM {
     pub fn arm_poll(&self) -> DmResult<DeviceInfo> {
         let mut hdr = DmOptions::default().to_ioctl_hdr(None, DmFlags::empty())?;
 
-        self.do_ioctl(dmi::DM_DEV_ARM_POLL_CMD as u8, &mut hdr, None)?;
-
-        DeviceInfo::new(hdr)
+        trace!("Issuing device-mapper arm poll command");
+        self.do_ioctl(dmi::DM_DEV_ARM_POLL_CMD as u8, &mut hdr, None)
+            .map(|(hdr, _)| hdr)
     }
 }
 
@@ -809,11 +852,14 @@ mod tests {
 
         let devices = dm.list_test_devices().unwrap();
 
-        assert_eq!(
-            devices.iter().map(|x| x.0.as_ref()).collect::<Vec<_>>(),
-            vec![&*name]
-        );
-        assert_eq!(devices[0].2.unwrap_or(0), 0);
+        assert_eq!(devices.len(), 1);
+
+        if dm.version().unwrap().1 >= 37 {
+            assert_matches!(devices.first().expect("len is 1"), (nm, _, Some(0)) if nm == &name);
+        } else {
+            assert_matches!(devices.first().expect("len is 1"), (nm, _, None) if nm == &name);
+        }
+
         dm.device_remove(&DevId::Name(&name), DmOptions::default())
             .unwrap();
     }
@@ -824,7 +870,10 @@ mod tests {
         let dm = DM::new().unwrap();
         let name = test_name("example-dev").expect("is valid DM name");
         let result = dm.device_create(&name, None, DmOptions::default()).unwrap();
+
         assert_eq!(result.name(), Some(&*name));
+        assert_eq!(result.uuid(), None);
+
         dm.device_remove(&DevId::Name(&name), DmOptions::default())
             .unwrap();
     }
@@ -838,8 +887,10 @@ mod tests {
         let result = dm
             .device_create(&name, Some(&uuid), DmOptions::default())
             .unwrap();
+
         assert_eq!(result.name(), Some(&*name));
-        assert_eq!(result.uuid().unwrap(), &*uuid);
+        assert_eq!(result.uuid(), Some(&*uuid));
+
         dm.device_remove(&DevId::Name(&name), DmOptions::default())
             .unwrap();
     }
@@ -857,7 +908,7 @@ mod tests {
 
         assert_matches!(
             dm.device_rename(&name, &DevId::Uuid(&new_uuid)),
-            Err(DmError::Core(Error::Ioctl(_, _)))
+            Err(DmError::Core(Error::Ioctl(op, _, _, err))) if *err == nix::errno::Errno::EINVAL && op == dmi::DM_DEV_RENAME_CMD as u8
         );
 
         dm.device_remove(&DevId::Name(&name), DmOptions::default())
@@ -866,6 +917,7 @@ mod tests {
 
     #[test]
     /// Verify that resetting uuid to same uuid fails.
+    /// Since a device with that UUID already exists, the UUID can not be used.
     fn sudo_test_rename_uuid_id() {
         let dm = DM::new().unwrap();
         let name = test_name("example-dev").expect("is valid DM name");
@@ -874,7 +926,7 @@ mod tests {
             .unwrap();
         assert_matches!(
             dm.device_rename(&name, &DevId::Uuid(&uuid)),
-            Err(DmError::Core(Error::Ioctl(_, _)))
+            Err(DmError::Core(Error::Ioctl(op, _, _, err))) if *err == nix::errno::Errno::EBUSY && op == dmi::DM_DEV_RENAME_CMD as u8
         );
 
         dm.device_remove(&DevId::Name(&name), DmOptions::default())
@@ -903,7 +955,7 @@ mod tests {
 
     #[test]
     /// Test that device rename to same name fails.
-    /// This is unfortunate, but appears to be true.
+    /// Since a device with that name already exists, the name can not be used.
     fn sudo_test_rename_id() {
         let dm = DM::new().unwrap();
         let name = test_name("example-dev").expect("is valid DM name");
@@ -911,7 +963,7 @@ mod tests {
 
         assert_matches!(
             dm.device_rename(&name, &DevId::Name(&name)),
-            Err(DmError::Core(Error::Ioctl(_, _)))
+            Err(DmError::Core(Error::Ioctl(op, _, _, err))) if *err == nix::errno::Errno::EBUSY && op == dmi::DM_DEV_RENAME_CMD as u8
         );
 
         dm.device_remove(&DevId::Name(&name), DmOptions::default())
@@ -932,14 +984,19 @@ mod tests {
 
         assert_matches!(
             dm.device_info(&DevId::Name(&name)),
-            Err(DmError::Core(Error::Ioctl(_, _)))
+            Err(DmError::Core(Error::Ioctl(_, _, _, err))) if *err == nix::errno::Errno::ENXIO
         );
 
         assert_matches!(dm.device_info(&DevId::Name(&new_name)), Ok(_));
 
         let devices = dm.list_test_devices().unwrap();
         assert_eq!(devices.len(), 1);
-        assert_eq!(devices[0].0.as_ref(), &*new_name);
+
+        if dm.version().unwrap().1 >= 37 {
+            assert_matches!(devices.first().expect("len is 1"), (nm, _, Some(0)) if nm == &new_name);
+        } else {
+            assert_matches!(devices.first().expect("len is 1"), (nm, _, None) if nm == &new_name);
+        }
 
         let third_name = test_name("example-dev-3").expect("is valid DM name");
         dm.device_create(&third_name, None, DmOptions::default())
@@ -947,7 +1004,7 @@ mod tests {
 
         assert_matches!(
             dm.device_rename(&new_name, &DevId::Name(&third_name)),
-            Err(DmError::Core(Error::Ioctl(_, _)))
+            Err(DmError::Core(Error::Ioctl(op, _, _, err))) if *err == nix::errno::Errno::EBUSY && op == dmi::DM_DEV_RENAME_CMD as u8
         );
 
         dm.device_remove(&DevId::Name(&third_name), DmOptions::default())
@@ -958,26 +1015,26 @@ mod tests {
 
     #[test]
     /// Renaming a device that does not exist yields an error.
-    fn sudo_test_rename_non_existant() {
+    fn sudo_test_rename_non_existent() {
         let new_name = test_name("new_name").expect("is valid DM name");
         assert_matches!(
             DM::new().unwrap().device_rename(
                 &test_name("old_name").expect("is valid DM name"),
                 &DevId::Name(&new_name)
             ),
-            Err(DmError::Core(Error::Ioctl(_, _)))
+            Err(DmError::Core(Error::Ioctl(op, _, _, err))) if *err == nix::errno::Errno::ENXIO && op == dmi::DM_DEV_RENAME_CMD as u8
         );
     }
 
     #[test]
-    /// Removing a device that does not exist yields an error, unfortunately.
-    fn sudo_test_remove_non_existant() {
+    /// Removing a device that does not exist yields an error.
+    fn sudo_test_remove_non_existent() {
         assert_matches!(
             DM::new().unwrap().device_remove(
                 &DevId::Name(&test_name("junk").expect("is valid DM name")),
                 DmOptions::default()
             ),
-            Err(DmError::Core(Error::Ioctl(_, _)))
+            Err(DmError::Core(Error::Ioctl(op, _, _, err))) if *err == nix::errno::Errno::ENXIO && op == dmi::DM_DEV_REMOVE_CMD as u8
         );
     }
 
@@ -997,27 +1054,27 @@ mod tests {
     }
 
     #[test]
-    /// Table status on a non-existant name should return an error.
-    fn sudo_test_table_status_non_existant() {
+    /// Table status on a non-existent name should return an error.
+    fn sudo_test_table_status_non_existent() {
         assert_matches!(
             DM::new().unwrap().table_status(
                 &DevId::Name(&test_name("junk").expect("is valid DM name")),
                 DmOptions::default()
             ),
-            Err(DmError::Core(Error::Ioctl(_, _)))
+            Err(DmError::Core(Error::Ioctl(_, _, _, err))) if *err == nix::errno::Errno::ENXIO
         );
     }
 
     #[test]
-    /// Table status on a non-existant name with TABLE_STATUS flag errors.
-    fn sudo_test_table_status_non_existant_table() {
+    /// Table status on a non-existent name with TABLE_STATUS flag errors.
+    fn sudo_test_table_status_non_existent_table() {
         let name = test_name("junk").expect("is valid DM name");
         assert_matches!(
             DM::new().unwrap().table_status(
                 &DevId::Name(&name),
                 DmOptions::default().set_flags(DmFlags::DM_STATUS_TABLE)
             ),
-            Err(DmError::Core(Error::Ioctl(_, _)))
+            Err(DmError::Core(Error::Ioctl(_, _, _, err))) if *err == nix::errno::Errno::ENXIO
         );
     }
 
@@ -1033,11 +1090,11 @@ mod tests {
         dm.device_create(&name, Some(&uuid), DmOptions::default())
             .unwrap();
 
-        let status = dm
+        let (hdr_out, status) = dm
             .table_status(&DevId::Name(&name), DmOptions::default())
             .unwrap();
-        assert!(status.1.is_empty());
-        assert_eq!(status.0.uuid(), Some(&*uuid));
+        assert!(status.is_empty());
+        assert_eq!(hdr_out.uuid(), Some(&*uuid));
         dm.device_remove(&DevId::Name(&name), DmOptions::default())
             .unwrap();
     }
@@ -1049,7 +1106,7 @@ mod tests {
         let name = test_name("example_dev").expect("is valid DM name");
         assert_matches!(
             DM::new().unwrap().device_info(&DevId::Name(&name)),
-            Err(DmError::Core(Error::Ioctl(_, _)))
+            Err(DmError::Core(Error::Ioctl(op, _, _, err))) if *err == nix::errno::Errno::ENXIO && op == dmi::DM_DEV_STATUS_CMD as u8
         );
     }
 
@@ -1068,19 +1125,19 @@ mod tests {
             .unwrap();
         assert_matches!(
             dm.device_create(&name, Some(&uuid), DmOptions::default()),
-            Err(DmError::Core(Error::Ioctl(_, _)))
+            Err(DmError::Core(Error::Ioctl(op, _, _, err))) if *err == nix::errno::Errno::EBUSY && op == dmi::DM_DEV_CREATE_CMD as u8
         );
         assert_matches!(
             dm.device_create(&name, None, DmOptions::default()),
-            Err(DmError::Core(Error::Ioctl(_, _)))
+            Err(DmError::Core(Error::Ioctl(op, _, _, err))) if *err == nix::errno::Errno::EBUSY && op == dmi::DM_DEV_CREATE_CMD as u8
         );
         assert_matches!(
             dm.device_create(&name, Some(&uuid_alt), DmOptions::default()),
-            Err(DmError::Core(Error::Ioctl(_, _)))
+            Err(DmError::Core(Error::Ioctl(op, _, _, err))) if *err == nix::errno::Errno::EBUSY && op == dmi::DM_DEV_CREATE_CMD as u8
         );
         assert_matches!(
             dm.device_create(&name_alt, Some(&uuid), DmOptions::default()),
-            Err(DmError::Core(Error::Ioctl(_, _)))
+            Err(DmError::Core(Error::Ioctl(op, _, _, err))) if *err == nix::errno::Errno::EBUSY && op == dmi::DM_DEV_CREATE_CMD as u8
         );
         dm.device_remove(&DevId::Name(&name), DmOptions::default())
             .unwrap();
