@@ -104,7 +104,7 @@ impl FromStr for Direction {
         } else if s == "w" {
             Ok(Direction::Writes)
         } else {
-            let err_msg = format!("Expected r or w, found {}", s);
+            let err_msg = format!("Expected r or w, found {s}");
             Err(DmError::Dm(ErrorEnum::Invalid, err_msg))
         }
     }
@@ -114,7 +114,7 @@ impl FromStr for Direction {
 /// If no feature parameters are present, during the periods of
 /// unreliability, all I/O returns errors.
 #[derive(Debug, Hash, Clone, Eq, PartialEq)]
-pub enum FeatureArg {
+pub enum FlakeyFeatureArg {
     /// drop_writes:
     ///
     /// All write I/O is silently ignored.
@@ -125,31 +125,29 @@ pub enum FeatureArg {
     /// All write I/O is failed with an error signalled.
     /// Read I/O is handled correctly.
     ErrorWrites,
-    /// corrupt_bio_byte <Nth_byte> <direction> <value> <flags>:
+    /// corrupt_bio_byte `<Nth_byte>` `<direction>` `<value>` `<flags>`:
     ///
-    /// During <down interval>, replace <Nth_byte> of the data of
-    /// each matching bio with <value>.
+    /// During `<down interval>`, replace `<Nth_byte>` of the data of
+    /// each matching bio with `<value>`.
     ///
-    /// <Nth_byte>: The offset of the byte to replace.
-    ///             Counting starts at 1, to replace the first byte.
-    /// <direction>: Either 'r' to corrupt reads or 'w' to corrupt writes.
-    ///             'w' is incompatible with drop_writes.
-    /// <value>:    The value (from 0-255) to write.
-    /// <flags>:    Perform the replacement only if bio->bi_opf has all the
-    ///             selected flags set.
+    /// * `<Nth_byte>`: The offset of the byte to replace.
+    ///   Counting starts at 1, to replace the first byte.
+    /// * `<direction>`: Either 'r' to corrupt reads or 'w' to corrupt writes.
+    ///   'w' is incompatible with drop_writes.
+    /// * `<value>`: The value (from 0-255) to write.
+    /// * `<flags>`: Perform the replacement only if `bio->bi_opf` has all the
+    ///   selected flags set.
     CorruptBioByte(u64, Direction, u8, u64),
 }
 
-impl fmt::Display for FeatureArg {
+impl fmt::Display for FlakeyFeatureArg {
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         match self {
-            FeatureArg::DropWrites => write!(f, "drop_writes"),
-            FeatureArg::ErrorWrites => write!(f, "error_writes"),
-            FeatureArg::CorruptBioByte(offset, direction, value, flags) => write!(
-                f,
-                "corrupt_bio_byte {} {} {} {}",
-                offset, direction, value, flags
-            ),
+            FlakeyFeatureArg::DropWrites => write!(f, "drop_writes"),
+            FlakeyFeatureArg::ErrorWrites => write!(f, "error_writes"),
+            FlakeyFeatureArg::CorruptBioByte(offset, direction, value, flags) => {
+                write!(f, "corrupt_bio_byte {offset} {direction} {value} {flags}")
+            }
         }
     }
 }
@@ -168,7 +166,7 @@ pub struct FlakeyTargetParams {
     /// DM source type is unsigned, so restrict to u32.
     pub down_interval: u32,
     /// Optional feature arguments
-    pub feature_args: HashSet<FeatureArg>,
+    pub feature_args: HashSet<FlakeyFeatureArg>,
 }
 
 impl FlakeyTargetParams {
@@ -178,7 +176,7 @@ impl FlakeyTargetParams {
         start_offset: Sectors,
         up_interval: u32,
         down_interval: u32,
-        feature_args: Vec<FeatureArg>,
+        feature_args: Vec<FlakeyFeatureArg>,
     ) -> FlakeyTargetParams {
         FlakeyTargetParams {
             device,
@@ -192,20 +190,18 @@ impl FlakeyTargetParams {
 
 impl fmt::Display for FlakeyTargetParams {
     /// Generate params to be passed to DM.  The format of the params is:
-    /// <dev path> <offset> <up interval> <down interval> \
-    ///   [<num_features> [<feature arguments>]]
     ///
-    /// Table parameters
-    /// ----------------
-    ///  <dev path> <offset> <up interval> <down interval> \
-    ///    [<num_features> [<feature arguments>]]
+    /// ```plain
+    /// <dev path> <offset> <up interval> <down interval> [<num_features> [<feature arguments>]]
+    /// ```
     ///
     /// Mandatory parameters:
-    ///    <dev path>: Full pathname to the underlying block-device, or a
-    ///                "major:minor" device-number.
-    ///    <offset>: Starting sector within the device.
-    ///    <up interval>: Number of seconds device is available.
-    ///    <down interval>: Number of seconds device returns errors.
+    ///
+    /// * `<dev path>`: Full pathname to the underlying block-device, or a
+    ///   "major:minor" device-number.
+    /// * `<offset>`: Starting sector within the device.
+    /// * `<up interval>`: Number of seconds device is available.
+    /// * `<down interval>`: Number of seconds device returns errors.
     ///
     /// Optional feature parameters:
     ///  If no feature parameters are present, during the periods of
@@ -219,13 +215,13 @@ impl FromStr for FlakeyTargetParams {
     type Err = DmError;
 
     fn from_str(s: &str) -> DmResult<FlakeyTargetParams> {
-        fn parse_feature_args(vals: &[&str]) -> DmResult<Vec<FeatureArg>> {
+        fn parse_feature_args(vals: &[&str]) -> DmResult<Vec<FlakeyFeatureArg>> {
             let mut vals_iter = vals.iter();
-            let mut result: Vec<FeatureArg> = Vec::new();
+            let mut result: Vec<FlakeyFeatureArg> = Vec::new();
             while let Some(x) = vals_iter.next() {
                 match x {
-                    &"drop_writes" => result.push(FeatureArg::DropWrites),
-                    &"error_writes" => result.push(FeatureArg::ErrorWrites),
+                    &"drop_writes" => result.push(FlakeyFeatureArg::DropWrites),
+                    &"error_writes" => result.push(FlakeyFeatureArg::ErrorWrites),
                     &"corrupt_bio_byte" => {
                         let offset = vals_iter
                             .next()
@@ -233,7 +229,7 @@ impl FromStr for FlakeyTargetParams {
                                 let err_msg = "corrupt_bio_byte takes 4 parameters";
                                 DmError::Dm(ErrorEnum::Invalid, err_msg.to_string())
                             })
-                            .and_then(|s| parse_value::<u64>(*s, "offset"))?;
+                            .and_then(|s| parse_value::<u64>(s, "offset"))?;
 
                         let direction = vals_iter
                             .next()
@@ -241,7 +237,7 @@ impl FromStr for FlakeyTargetParams {
                                 let err_msg = "corrupt_bio_byte takes 4 parameters";
                                 DmError::Dm(ErrorEnum::Invalid, err_msg.to_string())
                             })
-                            .and_then(|s| parse_value::<Direction>(*s, "direction"))?;
+                            .and_then(|s| parse_value::<Direction>(s, "direction"))?;
 
                         let value = vals_iter
                             .next()
@@ -249,7 +245,7 @@ impl FromStr for FlakeyTargetParams {
                                 let err_msg = "corrupt_bio_byte takes 4 parameters";
                                 DmError::Dm(ErrorEnum::Invalid, err_msg.to_string())
                             })
-                            .and_then(|s| parse_value::<u8>(*s, "value"))?;
+                            .and_then(|s| parse_value::<u8>(s, "value"))?;
 
                         let flags = vals_iter
                             .next()
@@ -257,12 +253,14 @@ impl FromStr for FlakeyTargetParams {
                                 let err_msg = "corrupt_bio_byte takes 4 parameters";
                                 DmError::Dm(ErrorEnum::Invalid, err_msg.to_string())
                             })
-                            .and_then(|s| parse_value::<u64>(*s, "flags"))?;
+                            .and_then(|s| parse_value::<u64>(s, "flags"))?;
 
-                        result.push(FeatureArg::CorruptBioByte(offset, direction, value, flags));
+                        result.push(FlakeyFeatureArg::CorruptBioByte(
+                            offset, direction, value, flags,
+                        ));
                     }
                     x => {
-                        let err_msg = format!("{} is an unrecognized feature parameter", x);
+                        let err_msg = format!("{x} is an unrecognized feature parameter");
                         return Err(DmError::Dm(ErrorEnum::Invalid, err_msg));
                     }
                 }
@@ -363,10 +361,10 @@ impl FromStr for LinearDevTargetParams {
     type Err = DmError;
 
     fn from_str(s: &str) -> DmResult<LinearDevTargetParams> {
-        let target_type = s.splitn(2, ' ').next().ok_or_else(|| {
+        let target_type = Some(s.split_once(' ').map_or(s, |x| x.0)).ok_or_else(|| {
             DmError::Dm(
                 ErrorEnum::Invalid,
-                format!("target line string \"{}\" did not contain any values", s),
+                format!("target line string \"{s}\" did not contain any values"),
             )
         })?;
         if target_type == FLAKEY_TARGET_NAME {
@@ -380,7 +378,7 @@ impl FromStr for LinearDevTargetParams {
         } else {
             Err(DmError::Dm(
                 ErrorEnum::Invalid,
-                format!("unexpected target type \"{}\"", target_type),
+                format!("unexpected target type \"{target_type}\""),
             ))
         }
     }
@@ -543,7 +541,7 @@ impl LinearDev {
             device_match(dm, &dev, uuid)?;
             dev
         } else {
-            let dev_info = device_create(dm, name, uuid, &table, DmOptions::default())?;
+            let dev_info = device_create(dm, name, uuid, &table, DmOptions::private())?;
             LinearDev {
                 dev_info: Box::new(dev_info),
                 table,
@@ -699,14 +697,8 @@ mod tests {
             .unwrap()
             .table;
         assert_eq!(table.len(), count);
-        match table[0].params {
-            LinearDevTargetParams::Linear(ref device) => assert_eq!(device.device, dev),
-            _ => panic!("unexpected param type"),
-        }
-        match table[1].params {
-            LinearDevTargetParams::Linear(ref device) => assert_eq!(device.device, dev),
-            _ => panic!("unexpected param type"),
-        }
+        assert_matches!(table[0].params, LinearDevTargetParams::Linear(ref device) if device.device == dev);
+        assert_matches!(table[1].params, LinearDevTargetParams::Linear(ref device) if device.device == dev);
 
         assert_eq!(
             blkdev_size(&OpenOptions::new().read(true).open(ld.devnode()).unwrap()).sectors(),
@@ -835,7 +827,7 @@ mod tests {
         let result = "flakey 8:32 0 16 2 1 drop_writes"
             .parse::<FlakeyTargetParams>()
             .unwrap();
-        let expected = [FeatureArg::DropWrites]
+        let expected = [FlakeyFeatureArg::DropWrites]
             .iter()
             .cloned()
             .collect::<HashSet<_>>();
@@ -847,7 +839,7 @@ mod tests {
         let result = "flakey 8:32 0 16 2 1 error_writes"
             .parse::<FlakeyTargetParams>()
             .unwrap();
-        let expected = [FeatureArg::ErrorWrites]
+        let expected = [FlakeyFeatureArg::ErrorWrites]
             .iter()
             .cloned()
             .collect::<HashSet<_>>();
@@ -859,7 +851,7 @@ mod tests {
         let result = "flakey 8:32 0 16 2 5 corrupt_bio_byte 32 r 1 0"
             .parse::<FlakeyTargetParams>()
             .unwrap();
-        let expected = [FeatureArg::CorruptBioByte(32, Direction::Reads, 1, 0)]
+        let expected = [FlakeyFeatureArg::CorruptBioByte(32, Direction::Reads, 1, 0)]
             .iter()
             .cloned()
             .collect::<HashSet<_>>();
@@ -871,10 +863,15 @@ mod tests {
         let result = "flakey 8:32 0 16 2 5 corrupt_bio_byte 224 w 0 32"
             .parse::<FlakeyTargetParams>()
             .unwrap();
-        let expected = [FeatureArg::CorruptBioByte(224, Direction::Writes, 0, 32)]
-            .iter()
-            .cloned()
-            .collect::<HashSet<_>>();
+        let expected = [FlakeyFeatureArg::CorruptBioByte(
+            224,
+            Direction::Writes,
+            0,
+            32,
+        )]
+        .iter()
+        .cloned()
+        .collect::<HashSet<_>>();
         assert_eq!(result.feature_args, expected);
     }
 
@@ -884,8 +881,8 @@ mod tests {
             .parse::<FlakeyTargetParams>()
             .unwrap();
         let expected = [
-            FeatureArg::CorruptBioByte(32, Direction::Reads, 1, 0),
-            FeatureArg::DropWrites,
+            FlakeyFeatureArg::CorruptBioByte(32, Direction::Reads, 1, 0),
+            FlakeyFeatureArg::DropWrites,
         ]
         .iter()
         .cloned()
@@ -899,8 +896,8 @@ mod tests {
             .parse::<FlakeyTargetParams>()
             .unwrap();
         let expected = [
-            FeatureArg::DropWrites,
-            FeatureArg::CorruptBioByte(32, Direction::Reads, 1, 0),
+            FlakeyFeatureArg::DropWrites,
+            FlakeyFeatureArg::CorruptBioByte(32, Direction::Reads, 1, 0),
         ]
         .iter()
         .cloned()
@@ -913,7 +910,7 @@ mod tests {
         let result = "flakey 8:32 0 16 2 2 error_writes drop_writes"
             .parse::<FlakeyTargetParams>()
             .unwrap();
-        let expected = [FeatureArg::ErrorWrites, FeatureArg::DropWrites]
+        let expected = [FlakeyFeatureArg::ErrorWrites, FlakeyFeatureArg::DropWrites]
             .iter()
             .cloned()
             .collect::<HashSet<_>>();
