@@ -5,7 +5,7 @@
 use std::{fmt, path::PathBuf, str::FromStr};
 
 use crate::{
-    core::{DevId, Device, DeviceInfo, DmCookie, DmFlags, DmName, DmOptions, DmUuid, DM},
+    core::{DevId, Device, DeviceInfo, DmFlags, DmName, DmOptions, DmUuid, DM},
     result::{DmError, DmResult, ErrorEnum},
     shared::{
         device_create, device_exists, device_match, get_status, get_status_line_fields, message,
@@ -18,14 +18,20 @@ use crate::{
 
 const THIN_TARGET_NAME: &str = "thin";
 
+/// Struct representing params for a thin target
 #[derive(Clone, Debug, Eq, PartialEq)]
 pub struct ThinTargetParams {
+    /// Thin pool for the given thin device
     pub pool: Device,
+    /// Thin ID
     pub thin_id: ThinDevId,
+    /// Optional block device outside of pool to be treated as a read-only snapshot
+    /// origin
     pub external_origin_dev: Option<Device>,
 }
 
 impl ThinTargetParams {
+    /// Create a new ThinTargetParams struct
     pub fn new(
         pool: Device,
         thin_id: ThinDevId,
@@ -52,10 +58,7 @@ impl FromStr for ThinTargetParams {
         let vals = s.split(' ').collect::<Vec<_>>();
         let len = vals.len();
         if !(3..=4).contains(&len) {
-            let err_msg = format!(
-                "expected 3 or 4 values in params string \"{}\", found {}",
-                s, len
-            );
+            let err_msg = format!("expected 3 or 4 values in params string \"{s}\", found {len}");
             return Err(DmError::Dm(ErrorEnum::Invalid, err_msg));
         }
 
@@ -95,12 +98,15 @@ impl TargetParams for ThinTargetParams {
     }
 }
 
+/// A target table for a thin device.
 #[derive(Clone, Debug, Eq, PartialEq)]
 pub struct ThinDevTargetTable {
+    /// The device's table
     pub table: TargetLine<ThinTargetParams>,
 }
 
 impl ThinDevTargetTable {
+    /// Make a new ThinDevTargetTable from required input
     pub fn new(start: Sectors, length: Sectors, params: ThinTargetParams) -> ThinDevTargetTable {
         ThinDevTargetTable {
             table: TargetLine::new(start, length, params),
@@ -163,6 +169,11 @@ impl DmDevice<ThinDevTargetTable> for ThinDev {
         name!(self)
     }
 
+    fn resume(&mut self, dm: &DM) -> DmResult<()> {
+        dm.device_suspend(&DevId::Name(self.name()), DmOptions::default())?;
+        Ok(())
+    }
+
     fn size(&self) -> Sectors {
         self.table.table.length
     }
@@ -172,7 +183,7 @@ impl DmDevice<ThinDevTargetTable> for ThinDev {
     }
 
     fn teardown(&mut self, dm: &DM) -> DmResult<()> {
-        dm.device_remove(&DevId::Name(self.name()), &DmOptions::new())?;
+        dm.device_remove(&DevId::Name(self.name()), DmOptions::default())?;
         Ok(())
     }
 
@@ -257,7 +268,7 @@ impl ThinDev {
         thin_pool: &ThinPoolDev,
         thin_id: ThinDevId,
     ) -> DmResult<ThinDev> {
-        message(dm, thin_pool, &format!("create_thin {}", thin_id))?;
+        message(dm, thin_pool, &format!("create_thin {thin_id}"))?;
 
         if device_exists(dm, name)? {
             let err_msg = "Uncreated device should not be known to kernel";
@@ -266,13 +277,7 @@ impl ThinDev {
 
         let thin_pool_device = thin_pool.device();
         let table = ThinDev::gen_default_table(length, thin_pool_device, thin_id);
-        let dev_info = device_create(
-            dm,
-            name,
-            uuid,
-            &table,
-            DmOptions::new().set_cookie(DmCookie::DM_UDEV_PRIMARY_SOURCE_FLAG),
-        )?;
+        let dev_info = device_create(dm, name, uuid, &table, DmOptions::default())?;
 
         Ok(ThinDev {
             dev_info: Box::new(dev_info),
@@ -308,13 +313,7 @@ impl ThinDev {
             device_match(dm, &dev, uuid)?;
             dev
         } else {
-            let dev_info = device_create(
-                dm,
-                name,
-                uuid,
-                &table,
-                DmOptions::new().set_cookie(DmCookie::DM_UDEV_PRIMARY_SOURCE_FLAG),
-            )?;
+            let dev_info = device_create(dm, name, uuid, &table, DmOptions::default())?;
             ThinDev {
                 dev_info: Box::new(dev_info),
                 table,
@@ -336,7 +335,10 @@ impl ThinDev {
         snapshot_thin_id: ThinDevId,
     ) -> DmResult<ThinDev> {
         let source_id = DevId::Name(self.name());
-        dm.device_suspend(&source_id, DmOptions::new().set_flags(DmFlags::DM_SUSPEND))?;
+        dm.device_suspend(
+            &source_id,
+            DmOptions::default().set_flags(DmFlags::DM_SUSPEND),
+        )?;
         message(
             dm,
             thin_pool,
@@ -345,23 +347,27 @@ impl ThinDev {
                 snapshot_thin_id, self.table.table.params.thin_id
             ),
         )?;
-        dm.device_suspend(&source_id, &DmOptions::new())?;
+        dm.device_suspend(&source_id, DmOptions::default())?;
         let table = ThinDev::gen_default_table(self.size(), thin_pool.device(), snapshot_thin_id);
         let dev_info = Box::new(device_create(
             dm,
             snapshot_name,
             snapshot_uuid,
             &table,
-            DmOptions::new().set_cookie(DmCookie::DM_UDEV_PRIMARY_SOURCE_FLAG),
+            DmOptions::default(),
         )?);
         Ok(ThinDev { dev_info, table })
     }
 
     /// Generate a table to be passed to DM. The format of the table
     /// entries is:
+    /// ```plain
     /// <start (0)> <length> "thin" <thin device specific string>
+    /// ```
     /// where the thin device specific string has the format:
+    /// ```plain
     /// <thinpool maj:min> <thin_id>
+    /// ```
     /// There is exactly one entry in the table.
     /// Various defaults are hard coded in the method.
     fn gen_default_table(
@@ -382,15 +388,15 @@ impl ThinDev {
     }
 
     /// Get the current status of the thin device.
-    pub fn status(&self, dm: &DM) -> DmResult<ThinStatus> {
-        status!(self, dm)
+    pub fn status(&self, dm: &DM, options: DmOptions) -> DmResult<ThinStatus> {
+        status!(self, dm, options)
     }
 
     /// Set the table for the thin device's target
     pub fn set_table(&mut self, dm: &DM, table: TargetLine<ThinTargetParams>) -> DmResult<()> {
         let table = ThinDevTargetTable::new(table.start, table.length, table.params);
-        self.suspend(dm, false)?;
-        self.table_load(dm, &table)?;
+        self.suspend(dm, DmOptions::default().set_flags(DmFlags::DM_NOFLUSH))?;
+        self.table_load(dm, &table, DmOptions::default())?;
         self.resume(dm)?;
 
         self.table = table;
@@ -402,7 +408,7 @@ impl ThinDev {
     pub fn destroy(&mut self, dm: &DM, thin_pool: &ThinPoolDev) -> DmResult<()> {
         let thin_id = self.table.table.params.thin_id;
         self.teardown(dm)?;
-        message(dm, thin_pool, &format!("delete {}", thin_id))?;
+        message(dm, thin_pool, &format!("delete {thin_id}"))?;
         Ok(())
     }
 }
@@ -411,7 +417,6 @@ impl ThinDev {
 mod tests {
 
     use std::{
-        collections::HashMap,
         fs::{canonicalize, OpenOptions},
         io::Write,
         path::Path,
@@ -422,7 +427,7 @@ mod tests {
 
     use crate::{
         consts::IEC,
-        core::errors::{Error, ErrorKind},
+        core::errors::Error,
         shared::DmDevice,
         testing::{
             blkdev_size, test_name, test_string, test_uuid, test_with_spec, udev_settle,
@@ -436,33 +441,6 @@ mod tests {
 
     const MIN_THIN_DEV_SIZE: Sectors = Sectors(1);
 
-    // Return a hashmap of key-value pairs for udev entry.
-    fn get_udev_db_entry(dev_node_search: &Path) -> Option<HashMap<String, String>> {
-        // Takes a libudev device entry and returns the properties as a HashMap.
-        fn device_as_map(device: &libudev::Device<'_>) -> HashMap<String, String> {
-            let rc: HashMap<_, _> = device
-                .properties()
-                .map(|i| {
-                    (
-                        String::from(i.name().to_str().unwrap()),
-                        String::from(i.value().to_str().unwrap()),
-                    )
-                })
-                .collect();
-            rc
-        }
-
-        let context = libudev::Context::new().unwrap();
-        let mut enumerator = libudev::Enumerator::new(&context).unwrap();
-        enumerator.match_subsystem("block").unwrap();
-
-        enumerator
-            .scan_devices()
-            .unwrap()
-            .find(|x| x.devnode().map_or(false, |d| dev_node_search == d))
-            .map(|dev| device_as_map(&dev))
-    }
-
     /// Verify that specifying a size of 0 Sectors will cause a failure.
     fn test_zero_size(paths: &[&Path]) {
         assert!(!paths.is_empty());
@@ -506,7 +484,7 @@ mod tests {
                 &tp,
                 ThinDevId::new_u64(0).expect("is below limit")
             ),
-            Err(DmError::Core(Error(ErrorKind::IoctlError(_, _), _)))
+            Err(DmError::Core(Error::Ioctl(_, _, _, _)))
         );
 
         tp.teardown(&dm).unwrap();
@@ -540,7 +518,7 @@ mod tests {
         assert_eq!(table.params.thin_id, thin_id);
 
         assert_matches!(
-            td.status(&dm).unwrap(),
+            td.status(&dm, DmOptions::default()).unwrap(),
             ThinStatus::Error | ThinStatus::Working(_)
         );
 
@@ -552,7 +530,7 @@ mod tests {
         // New thindev w/ same id fails.
         assert_matches!(
             ThinDev::new(&dm, &id, None, td_size, &tp, thin_id),
-            Err(DmError::Core(Error(ErrorKind::IoctlError(_, _), _)))
+            Err(DmError::Core(Error::Ioctl(_, _, _, _)))
         );
 
         // Verify that the device of that name does exist.
@@ -577,25 +555,15 @@ mod tests {
     /// Test thin device create, load, and snapshot and make sure that all is well with udev
     /// db and symlink generation.
     fn test_udev_userspace(paths: &[&Path]) {
-        // Make sure we are meeting all our expectations in user space with regards to udev
-        // handling.
+        // Confirm that the correct symlink has been constructed.
         fn validate(path_uuid: &Uuid, devnode: &Path) {
             udev_settle().unwrap();
 
             // Make sure the uuid symlink was created
-            let symlink = PathBuf::from(format!("/dev/disk/by-uuid/{}", path_uuid));
-
+            let symlink = PathBuf::from(format!("/dev/disk/by-uuid/{path_uuid}"));
             assert!(symlink.exists());
 
-            // Make sure the symlink points to devnode
-            let uuid_sym = canonicalize(symlink).unwrap();
-
-            assert_eq!(*devnode, uuid_sym);
-
-            // Check the udev db to see what it's showing for the device, make sure SYSTEMD_READY=0
-            // is not present
-            let entry = get_udev_db_entry(devnode).unwrap();
-            assert!(!entry.contains_key("SYSTEMD_READY"));
+            assert_eq!(*devnode, canonicalize(symlink).unwrap());
         }
 
         // Set the FS with devnode to a new auto generated UUID
@@ -629,11 +597,13 @@ mod tests {
         let mut td = ThinDev::new(&dm, &id, None, tp.size(), &tp, thin_id).unwrap();
         udev_settle().unwrap();
 
-        // Create the XFS FS on top of the thin device
-        xfs_create_fs(&td.devnode()).unwrap();
+        let uuid = Uuid::new_v4();
 
-        // Travis CI is old doesn't support setting uuid during FS creation.
-        let uuid = set_new_fs_uuid(&td.devnode());
+        // Create the XFS FS on top of the thin device
+        xfs_create_fs(&td.devnode(), Some(uuid)).unwrap();
+
+        // Synchronize with udev processing triggered by xfs_create_fs()
+        udev_settle().unwrap();
 
         validate(&uuid, &td.devnode());
 
@@ -669,7 +639,7 @@ mod tests {
         let dm = DM::new().unwrap();
         let mut tp = minimal_thinpool(&dm, paths[0]);
 
-        let orig_data_usage = match tp.status(&dm).unwrap() {
+        let orig_data_usage = match tp.status(&dm, DmOptions::default()).unwrap() {
             ThinPoolStatus::Working(ref status) => status.usage.used_data,
             ThinPoolStatus::Error => panic!("devicemapper could not obtain thin pool status"),
             ThinPoolStatus::Fail => panic!("failed to get thinpool status"),
@@ -683,7 +653,7 @@ mod tests {
         let mut td = ThinDev::new(&dm, &thin_name, None, td_size, &tp, thin_id).unwrap();
         udev_settle().unwrap();
 
-        let data_usage_1 = match tp.status(&dm).unwrap() {
+        let data_usage_1 = match tp.status(&dm, DmOptions::default()).unwrap() {
             ThinPoolStatus::Working(ref status) => status.usage.used_data,
             ThinPoolStatus::Error => panic!("devicemapper could not obtain thin pool status"),
             ThinPoolStatus::Fail => panic!("failed to get thinpool status"),
@@ -697,7 +667,7 @@ mod tests {
         let mut ss = td.snapshot(&dm, &ss_name, None, &tp, ss_id).unwrap();
         udev_settle().unwrap();
 
-        let data_usage_2 = match tp.status(&dm).unwrap() {
+        let data_usage_2 = match tp.status(&dm, DmOptions::default()).unwrap() {
             ThinPoolStatus::Working(ref status) => status.usage.used_data,
             ThinPoolStatus::Error => panic!("devicemapper could not obtain thin pool status"),
             ThinPoolStatus::Fail => panic!("failed to get thinpool status"),
@@ -727,16 +697,16 @@ mod tests {
         let mut td = ThinDev::new(&dm, &thin_name, None, tp.size(), &tp, thin_id).unwrap();
         udev_settle().unwrap();
 
-        let orig_data_usage = match tp.status(&dm).unwrap() {
+        let orig_data_usage = match tp.status(&dm, DmOptions::default()).unwrap() {
             ThinPoolStatus::Working(ref status) => status.usage.used_data,
             ThinPoolStatus::Error => panic!("devicemapper could not obtain thin pool status"),
             ThinPoolStatus::Fail => panic!("failed to get thinpool status"),
         };
         assert_eq!(orig_data_usage, DataBlocks(0));
 
-        xfs_create_fs(&td.devnode()).unwrap();
+        xfs_create_fs(&td.devnode(), None).unwrap();
 
-        let data_usage_1 = match tp.status(&dm).unwrap() {
+        let data_usage_1 = match tp.status(&dm, DmOptions::default()).unwrap() {
             ThinPoolStatus::Working(ref status) => status.usage.used_data,
             ThinPoolStatus::Error => panic!("devicemapper could not obtain thin pool status"),
             ThinPoolStatus::Fail => panic!("failed to get thinpool status"),
@@ -757,10 +727,11 @@ mod tests {
         .unwrap();
 
         for i in 0..100 {
-            let file_path = tmp_dir.path().join(format!("devicemapper_test{}.txt", i));
+            let file_path = tmp_dir.path().join(format!("devicemapper_test{i}.txt"));
             writeln!(
                 &OpenOptions::new()
                     .create(true)
+                    .truncate(true)
                     .write(true)
                     .open(file_path)
                     .unwrap(),
@@ -770,7 +741,7 @@ mod tests {
         }
         umount2(tmp_dir.path(), MntFlags::MNT_DETACH).unwrap();
 
-        let data_usage_2 = match tp.status(&dm).unwrap() {
+        let data_usage_2 = match tp.status(&dm, DmOptions::default()).unwrap() {
             ThinPoolStatus::Working(ref status) => status.usage.used_data,
             ThinPoolStatus::Error => panic!("devicemapper could not obtain thin pool status"),
             ThinPoolStatus::Fail => panic!("failed to get thinpool status"),
@@ -801,16 +772,16 @@ mod tests {
             ThinDev::new(&dm, &thin_name, None, Sectors(2 * IEC::Mi), &tp, thin_id).unwrap();
         udev_settle().unwrap();
 
-        let orig_data_usage = match tp.status(&dm).unwrap() {
+        let orig_data_usage = match tp.status(&dm, DmOptions::default()).unwrap() {
             ThinPoolStatus::Working(ref status) => status.usage.used_data,
             ThinPoolStatus::Error => panic!("devicemapper could not obtain thin pool status"),
             ThinPoolStatus::Fail => panic!("failed to get thinpool status"),
         };
         assert_eq!(orig_data_usage, DataBlocks(0));
 
-        xfs_create_fs(&td.devnode()).unwrap();
+        xfs_create_fs(&td.devnode(), None).unwrap();
 
-        let data_usage_1 = match tp.status(&dm).unwrap() {
+        let data_usage_1 = match tp.status(&dm, DmOptions::default()).unwrap() {
             ThinPoolStatus::Working(ref status) => status.usage.used_data,
             ThinPoolStatus::Error => panic!("devicemapper could not obtain thin pool status"),
             ThinPoolStatus::Fail => panic!("failed to get thinpool status"),
@@ -826,7 +797,7 @@ mod tests {
             .unwrap();
         udev_settle().unwrap();
 
-        let data_usage_2 = match tp.status(&dm).unwrap() {
+        let data_usage_2 = match tp.status(&dm, DmOptions::default()).unwrap() {
             ThinPoolStatus::Working(ref status) => status.usage.used_data,
             ThinPoolStatus::Error => panic!("devicemapper could not obtain thin pool status"),
             ThinPoolStatus::Fail => panic!("failed to get thinpool status"),
@@ -838,7 +809,7 @@ mod tests {
         // Setting the uuid of the snapshot filesystem bumps the usage,
         // but does not increase the usage quite as much as establishing
         // the origin.
-        let data_usage_3 = match tp.status(&dm).unwrap() {
+        let data_usage_3 = match tp.status(&dm, DmOptions::default()).unwrap() {
             ThinPoolStatus::Working(ref status) => status.usage.used_data,
             ThinPoolStatus::Error => panic!("devicemapper could not obtain thin pool status"),
             ThinPoolStatus::Fail => panic!("failed to get thinpool status"),
@@ -853,16 +824,16 @@ mod tests {
             ThinDev::new(&dm, &thin_name, None, Sectors(2 * IEC::Gi), &tp, thin_id).unwrap();
         udev_settle().unwrap();
 
-        let data_usage_4 = match tp.status(&dm).unwrap() {
+        let data_usage_4 = match tp.status(&dm, DmOptions::default()).unwrap() {
             ThinPoolStatus::Working(ref status) => status.usage.used_data,
             ThinPoolStatus::Error => panic!("devicemapper could not obtain thin pool status"),
             ThinPoolStatus::Fail => panic!("failed to get thinpool status"),
         };
         assert_eq!(data_usage_4, data_usage_3);
 
-        xfs_create_fs(&td1.devnode()).unwrap();
+        xfs_create_fs(&td1.devnode(), None).unwrap();
 
-        let data_usage_5 = match tp.status(&dm).unwrap() {
+        let data_usage_5 = match tp.status(&dm, DmOptions::default()).unwrap() {
             ThinPoolStatus::Working(ref status) => status.usage.used_data,
             ThinPoolStatus::Error => panic!("devicemapper could not obtain thin pool status"),
             ThinPoolStatus::Fail => panic!("failed to get thinpool status"),
@@ -897,7 +868,7 @@ mod tests {
         // This should fail
         assert_matches!(
             ThinDev::setup(&dm, &thin_name, None, tp.size(), &tp, thin_id),
-            Err(DmError::Core(Error(ErrorKind::IoctlError(_, _), _)))
+            Err(DmError::Core(Error::Ioctl(_, _, _, _)))
         );
 
         tp.teardown(&dm).unwrap();
