@@ -24,16 +24,23 @@ use crate::core::devnode_to_devno;
 
 const THINPOOL_TARGET_NAME: &str = "thin-pool";
 
+/// Struct representing params for a thin pool target
 #[derive(Clone, Debug, Eq, PartialEq)]
 pub struct ThinPoolTargetParams {
+    /// Thin pool metadata device
     pub metadata_dev: Device,
+    /// Thin pool data device
     pub data_dev: Device,
+    /// Block size for allocations within the thin pool
     pub data_block_size: Sectors,
+    /// Amount of free space left at which to trigger the low water mark
     pub low_water_mark: DataBlocks,
+    /// Feature arguments
     pub feature_args: HashSet<String>,
 }
 
 impl ThinPoolTargetParams {
+    /// Create a new ThinPoolTargetParams struct
     pub fn new(
         metadata_dev: Device,
         data_dev: Device,
@@ -136,12 +143,15 @@ impl TargetParams for ThinPoolTargetParams {
     }
 }
 
+/// A target table for a thin pool device.
 #[derive(Clone, Debug, Eq, PartialEq)]
 pub struct ThinPoolDevTargetTable {
+    /// The device's table
     pub table: TargetLine<ThinPoolTargetParams>,
 }
 
 impl ThinPoolDevTargetTable {
+    /// Make a new ThinPoolDevTargetTable from a suitable vec
     pub fn new(
         start: Sectors,
         length: Sectors,
@@ -210,12 +220,12 @@ impl DmDevice<ThinPoolDevTargetTable> for ThinPoolDev {
     ) -> DmResult<bool> {
         let left = &left.table;
         let right = &right.table;
+
         Ok(left.start == right.start
             && left.length == right.length
             && left.params.metadata_dev == right.params.metadata_dev
             && left.params.data_dev == right.params.data_dev
-            && left.params.data_block_size == right.params.data_block_size
-            && left.params.feature_args == right.params.feature_args)
+            && left.params.data_block_size == right.params.data_block_size)
     }
 
     fn name(&self) -> &DmName {
@@ -397,7 +407,7 @@ impl FromStr for ThinPoolStatus {
             "-" => false,
             "needs_check" => true,
             val => {
-                return Err(make_unexpected_value_error(8, val, "needs checK"));
+                return Err(make_unexpected_value_error(8, val, "needs check"));
             }
         };
 
@@ -423,13 +433,13 @@ impl FromStr for ThinPoolStatus {
 /// Use DM to create a "thin-pool".  A "thin-pool" is shared space for
 /// other thin provisioned devices to use.
 ///
-/// See section "Setting up a fresh pool device":
-/// https://www.kernel.org/doc/Documentation/device-mapper/thin-provisioning.txt
+/// See section ["Setting up a fresh pool device"](https://docs.kernel.org/admin-guide/device-mapper/thin-provisioning.html#setting-up-a-fresh-pool-device)
 impl ThinPoolDev {
-    /// Construct a new ThinPoolDev with the given data and meta devs.
+    /// Construct a new `ThinPoolDev` with the given data and meta devs.
     /// Returns an error if the device is already known to the kernel.
-    /// Returns an error if data_block_size is not within required range.
+    /// Returns an error if `data_block_size` is not within required range.
     /// Precondition: the metadata device does not contain any pool metadata.
+    #[allow(clippy::too_many_arguments)]
     pub fn new(
         dm: &DM,
         name: &DmName,
@@ -438,14 +448,16 @@ impl ThinPoolDev {
         data: LinearDev,
         data_block_size: Sectors,
         low_water_mark: DataBlocks,
+        feature_args: Vec<String>,
     ) -> DmResult<ThinPoolDev> {
         if device_exists(dm, name)? {
-            let err_msg = format!("thinpooldev {} already exists", name);
+            let err_msg = format!("thinpooldev {name} already exists");
             return Err(DmError::Dm(ErrorEnum::Invalid, err_msg));
         }
 
-        let table = ThinPoolDev::gen_default_table(&meta, &data, data_block_size, low_water_mark);
-        let dev_info = device_create(dm, name, uuid, &table, DmOptions::default())?;
+        let table =
+            ThinPoolDev::gen_table(&meta, &data, data_block_size, low_water_mark, feature_args);
+        let dev_info = device_create(dm, name, uuid, &table, DmOptions::private())?;
 
         Ok(ThinPoolDev {
             dev_info: Box::new(dev_info),
@@ -500,7 +512,7 @@ impl ThinPoolDev {
             device_match(dm, &dev, uuid)?;
             dev
         } else {
-            let dev_info = device_create(dm, name, uuid, &table, DmOptions::default())?;
+            let dev_info = device_create(dm, name, uuid, &table, DmOptions::private())?;
             ThinPoolDev {
                 dev_info: Box::new(dev_info),
                 meta_dev: meta,
@@ -513,9 +525,13 @@ impl ThinPoolDev {
 
     /// Generate a table to be passed to DM. The format of the table
     /// entries is:
+    /// ```plain
     /// <start sec (0)> <length> "thin-pool" <thin-pool-specific string>
+    /// ```
     /// where the thin-pool-specific string has the format:
+    /// ```plain
     /// <meta maj:min> <data maj:min> <block size> <low water mark>
+    /// ```
     /// There is exactly one entry in the table.
     fn gen_table(
         meta: &LinearDev,
@@ -537,31 +553,6 @@ impl ThinPoolDev {
         )
     }
 
-    /// Generate a table to be passed to DM. The format of the table
-    /// entries is:
-    /// <start sec (0)> <length> "thin-pool" <thin-pool-specific string>
-    /// where the thin-pool-specific string has the format:
-    /// <meta maj:min> <data maj:min> <block size> <low water mark>
-    /// There is exactly one entry in the table.
-    /// Various defaults are hard coded in the method.
-    fn gen_default_table(
-        meta: &LinearDev,
-        data: &LinearDev,
-        data_block_size: Sectors,
-        low_water_mark: DataBlocks,
-    ) -> ThinPoolDevTargetTable {
-        Self::gen_table(
-            meta,
-            data,
-            data_block_size,
-            low_water_mark,
-            vec![
-                "no_discard_passdown".to_owned(),
-                "skip_block_zeroing".to_owned(),
-            ],
-        )
-    }
-
     /// Set the low water mark.
     /// This action puts the device in a state where it is ready to be resumed.
     pub fn set_low_water_mark(&mut self, dm: &DM, low_water_mark: DataBlocks) -> DmResult<()> {
@@ -628,29 +619,31 @@ impl ThinPoolDev {
         Ok(())
     }
 
-    /// Default behavior for devicemapper thin pools is to queue requests if
-    /// the thin pool is out of space to allow time for the thin pool to extend.
-    /// This behavior can be changed by adding the feature argument
-    /// `error_if_no_space` to the devicemapper table.
-    ///
-    /// This method will add `error_if_no_space` from the devicemapper table
-    /// if it is not present.
-    pub fn error_if_no_space(&mut self, dm: &DM) -> DmResult<()> {
-        self.suspend(dm, DmOptions::default().set_flags(DmFlags::DM_NOFLUSH))?;
-
+    fn set_feature_arg(&mut self, feature_arg: &str, dm: &DM) -> DmResult<()> {
         let mut table = self.table().clone();
-        if !table
-            .table
-            .params
-            .feature_args
-            .contains("error_if_no_space")
-        {
+        if !table.table.params.feature_args.contains(feature_arg) {
             table
                 .table
                 .params
                 .feature_args
-                .insert("error_if_no_space".to_string());
+                .insert(feature_arg.to_string());
 
+            self.suspend(dm, DmOptions::default().set_flags(DmFlags::DM_NOFLUSH))?;
+            self.table_load(dm, &table, DmOptions::default())?;
+            self.table = table;
+
+            self.resume(dm)?;
+        }
+
+        Ok(())
+    }
+
+    fn unset_feature_arg(&mut self, feature_arg: &str, dm: &DM) -> DmResult<()> {
+        let mut table = self.table().clone();
+        if table.table.params.feature_args.contains(feature_arg) {
+            table.table.params.feature_args.remove(feature_arg);
+
+            self.suspend(dm, DmOptions::default().set_flags(DmFlags::DM_NOFLUSH))?;
             self.table_load(dm, &table, DmOptions::default())?;
             self.table = table;
 
@@ -660,6 +653,17 @@ impl ThinPoolDev {
         Ok(())
     }
 
+    /// Default behavior for devicemapper thin pools is to queue requests if
+    /// the thin pool is out of space to allow time for the thin pool to extend.
+    /// This behavior can be changed by adding the feature argument
+    /// `error_if_no_space` to the devicemapper table.
+    ///
+    /// This method will add `error_if_no_space` from the devicemapper table
+    /// if it is not present.
+    pub fn error_if_no_space(&mut self, dm: &DM) -> DmResult<()> {
+        self.set_feature_arg("error_if_no_space", dm)
+    }
+
     /// Default behavior for devicemapper thin pools is to queue requests if
     /// the thin pool is out of space to allow time for the thin pool to extend.
     /// This behavior can be changed by adding the feature argument
@@ -668,24 +672,47 @@ impl ThinPoolDev {
     /// This method will remove `error_if_no_space` from the devicemapper table
     /// if it is present.
     pub fn queue_if_no_space(&mut self, dm: &DM) -> DmResult<()> {
-        self.suspend(dm, DmOptions::default().set_flags(DmFlags::DM_NOFLUSH))?;
+        self.unset_feature_arg("error_if_no_space", dm)
+    }
 
-        let mut table = self.table().clone();
-        if table
-            .table
-            .params
-            .feature_args
-            .contains("error_if_no_space")
-        {
-            table.table.params.feature_args.remove("error_if_no_space");
+    /// Default behavior for devicemapper thin pools is to zero newly allocated
+    /// data blocks. This behavior can be changed by adding the feature argument
+    /// `skip_block_zeroing` to the devicemapper table.
+    ///
+    /// This method will add `skip_block_zeroing` from the devicemapper table
+    /// if it is not present.
+    pub fn skip_block_zeroing(&mut self, dm: &DM) -> DmResult<()> {
+        self.set_feature_arg("skip_block_zeroing", dm)
+    }
 
-            self.table_load(dm, &table, DmOptions::default())?;
-            self.table = table;
+    /// Default behavior for devicemapper thin pools is to zero newly allocated
+    /// data blocks. This behavior can be changed by adding the feature argument
+    /// `skip_block_zeroing` to the devicemapper table.
+    ///
+    /// This method will remove `skip_block_zeroing` from the devicemapper table
+    /// if it is present.
+    pub fn require_block_zeroing(&mut self, dm: &DM) -> DmResult<()> {
+        self.unset_feature_arg("skip_block_zeroing", dm)
+    }
 
-            self.resume(dm)?;
-        }
+    /// Default behavior for devicemapper thin pools is to pass down discards.
+    /// This behavior can be changed by adding the feature argument
+    /// `no_discard_passdown` to the devicemapper table.
+    ///
+    /// This method will add `no_discard_passdown` to the devicemapper table
+    /// if it is not present.
+    pub fn no_discard_passdown(&mut self, dm: &DM) -> DmResult<()> {
+        self.set_feature_arg("no_discard_passdown", dm)
+    }
 
-        Ok(())
+    /// Default behavior for devicemapper thin pools is to pass down discards.
+    /// This behavior can be changed by adding the feature argument
+    /// `no_discard_passdown` to the devicemapper table.
+    ///
+    /// This method will remove `no_discard_passdown` from the devicemapper
+    /// table if it is present.
+    pub fn discard_passdown(&mut self, dm: &DM) -> DmResult<()> {
+        self.unset_feature_arg("no_discard_passdown", dm)
     }
 }
 
@@ -756,6 +783,10 @@ pub fn minimal_thinpool(dm: &DM, path: &Path) -> ThinPoolDev {
         data,
         MIN_DATA_BLOCK_SIZE,
         DataBlocks(1),
+        vec![
+            "no_discard_passdown".to_owned(),
+            "skip_block_zeroing".to_owned(),
+        ],
     )
     .unwrap()
 }
@@ -796,7 +827,7 @@ mod tests {
                     DataBlocks(tp.data_dev().size() / tp.data_block_size())
                 );
             }
-            status => panic!("unexpected thinpool status: {:?}", status),
+            status => panic!("unexpected thinpool status: {status:?}"),
         }
 
         let table = ThinPoolDev::read_kernel_table(&dm, &DevId::Name(tp.name()))
@@ -847,9 +878,13 @@ mod tests {
                 meta,
                 data,
                 MIN_DATA_BLOCK_SIZE / 2u64,
-                DataBlocks(1)
+                DataBlocks(1),
+                vec![
+                    "no_discard_passdown".to_owned(),
+                    "skip_block_zeroing".to_owned()
+                ],
             ),
-            Err(DmError::Core(Error::Ioctl(_, _)))
+            Err(DmError::Core(Error::Ioctl(_, _, _, _)))
         );
         dm.device_remove(&DevId::Name(&meta_name), DmOptions::default())
             .unwrap();
